/*
 videoTools - R package; display raw frame buffer on Mac OS X
 
 (C)Copyright 2007 Simon Urbanek <urbanek@r-project.org>
 
 Based on vo_macosx MPlayer Mac OSX video out module by 
 Nicolas Plourde <nicolasplourde@gmail.com> Copyright (c) 2005
 
 License: GPL v2
*/

#ifndef __APPLE__
/* this module works on Mac OS X only */
#else

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>

#include <R.h>
#include <Rinternals.h>

#import "macdisplay.h"

#define IMGFMT_BGR32 0x52474220
#define IMGFMT_RGB32 0x42475220
#define IMGFMT_YUY2  0x32595559
#define IMGFMT_YV12  0x32315659
#define IMGFMT_IYUV  0x56555949
#define IMGFMT_RGB24 0x42475218
#define IMGFMT_BGR24 0x52474218

//Cocoa
NSAutoreleasePool *autoreleasepool;

static BOOL initialized = NO;
static int ids = 0;

static void uninit(void)
{
/*
		[mplayerosxProxy stop];
		if (shmdt(image_data) == -1)
			mp_msg(MSGT_VO, MSGL_FATAL, "uninit: shmdt failed\n");
		if (shmctl(shm_id, IPC_RMID, &shm_desc) == -1)
			mp_msg(MSGT_VO, MSGL_FATAL, "uninit: shmctl failed\n"); */

    SetSystemUIMode( kUIModeNormal, 0);
    CGDisplayShowCursor(kCGDirectMainDisplay);
    
    if(autoreleasepool)
        [autoreleasepool release];
	autoreleasepool=0;
}

static int preinit(const char *arg)
{
	int parse_err = 0;

	NSApplicationLoad();
	autoreleasepool = [[NSAutoreleasePool alloc] init];
	NSApp = [NSApplication sharedApplication];
	
#if !defined (MACOSX_FINDER_SUPPORT) || !defined (HAVE_SDL)
	{
		//this chunk of code is heavily based off SDL_macosx.m from SDL 
		//it uses an Apple private function to request foreground operation
		void CPSEnableForegroundOperation(ProcessSerialNumber* psn);
		ProcessSerialNumber myProc, frProc;
		Boolean sameProc;

		if (GetFrontProcess(&frProc) == noErr)
		{
			if (GetCurrentProcess(&myProc) == noErr)
			{
				if (SameProcess(&frProc, &myProc, &sameProc) == noErr && !sameProc)
				{
					CPSEnableForegroundOperation(&myProc);
				}
				SetFrontProcess(&myProc);
			}
		}
	}
#endif
    return 0;
}

@implementation VideoView
- (int) configureWidth: (uint32_t) width height: (uint32_t) height format: (uint32_t) format
{
	int i;
	
	//init screen
	NSArray *screen_array = [NSScreen screens];
	screen_handle = [screen_array objectAtIndex:0];
	screen_frame = [screen_handle frame];

	image_format = format;
	image_width  = width;
	image_height = height;
	switch (image_format) 
	{
		case IMGFMT_BGR32:
		case IMGFMT_RGB32:
			image_depth = 32;
			pixelFormat = k32ARGBPixelFormat;
			break;
		case IMGFMT_YUY2:
			image_depth = 16;
			pixelFormat = kYUVSPixelFormat;
			break;
	}
	
	ids++;
	did = ids;
	
	image_bytes = (image_depth + 7) / 8;
	image_data = malloc(image_width*image_height*image_bytes);
	fprintf(stderr, "image_bytes=%d, pixelFormat=%x [%x], image_depth=%d, %d x %d\n", image_bytes, pixelFormat, format, image_depth, image_width, image_height);

	//init menu
	[self initMenu];
	
	//create window
	window = [[NSWindow alloc]	initWithContentRect:NSMakeRect(0, 0, 100, 100) styleMask:NSTitledWindowMask|NSTexturedBackgroundWindowMask|NSClosableWindowMask|NSMiniaturizableWindowMask|NSResizableWindowMask backing:NSBackingStoreBuffered defer:NO];
	
	[window autorelease];
	[window setDelegate:self];
	[window setContentView:self];
	[window setInitialFirstResponder:self];
	[window setAcceptsMouseMovedEvents:YES];
    [window setTitle:@"Video"];
	
	isFullscreen = 0;
	winSizeMult = 1;
	monitor_aspect = 1;
	movie_aspect = 1;
	isRootwin = 0;
	vo_keepaspect = 0;

	[self config];
	
	return 0;
}

- (id) config
{
	uint32_t d_width = image_width;
	uint32_t d_height = image_height;
	
	long swapInterval = 1;
	
	NSRect frame;
	CVReturn rerror = kCVReturnSuccess;
	
	//config window
	frame = NSMakeRect(0, 0, d_width, d_height);
	[window setContentSize: frame.size];
	
	//create OpenGL Context
	glContext = [[NSOpenGLContext alloc] initWithFormat:[NSOpenGLView defaultPixelFormat] shareContext:nil];	
	
	[self setOpenGLContext:glContext];
	[glContext setValues:&swapInterval forParameter:NSOpenGLCPSwapInterval];
	[glContext setView:self];
	[glContext makeCurrentContext];	
	
	rerror = CVPixelBufferCreateWithBytes( NULL, image_width, image_height, pixelFormat, image_data, image_width*image_bytes, NULL, NULL, NULL, &currentFrameBuffer);
	if(rerror != kCVReturnSuccess)
		error("Failed to create Pixel Buffer(%d)\n", rerror);
	
	rerror = CVOpenGLTextureCacheCreate(NULL, 0, [glContext CGLContextObj], [[self pixelFormat] CGLPixelFormatObj], 0, &textureCache);
	if(rerror != kCVReturnSuccess)
		error("Failed to create OpenGL texture Cache(%d)\n", rerror);
	
	rerror = CVOpenGLTextureCacheCreateTextureFromImage(	NULL, textureCache, currentFrameBuffer, 0, &texture);
	if(rerror != kCVReturnSuccess)
		error("Failed to create OpenGL texture(%d)\n", rerror);
	
	//show window
	[window center];
	[window makeKeyAndOrderFront:self];
	
	//if(vo_rootwin)
	//	[mpGLView rootwin];	

	//if(vo_fs)
	//	[mpGLView fullscreen: NO];
	
	//if(vo_ontop)
	//	[mpGLView ontop];
	return self;
}

/*
	Init Menu
*/
- (void)initMenu
{
	NSMenu *menu, *aspectMenu;
	NSMenuItem *menuItem;
	
	BOOL soleMenu = (![NSApp mainMenu]);
	if (soleMenu) [NSApp setMainMenu:[[NSMenu alloc] init]];

	// Create Movie Menu (but only if it's not there already)
	if ([[NSApp mainMenu] indexOfItemWithTitle:@"Movie"]<0) {
		menu = [[NSMenu alloc] initWithTitle:@"Movie"];
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Half Size" action:@selector(menuAction:) keyEquivalent:@""]; [menu addItem:menuItem];
		kHalfScreenCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Normal Size" action:@selector(menuAction:) keyEquivalent:@"1"];
		if (!soleMenu) [menuItem setKeyEquivalentModifierMask: NSCommandKeyMask|NSAlternateKeyMask];
		[menu addItem:menuItem];
		kNormalScreenCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Double Size" action:@selector(menuAction:) keyEquivalent:@"2"]; [menu addItem:menuItem];
		if (!soleMenu) [menuItem setKeyEquivalentModifierMask: NSCommandKeyMask|NSAlternateKeyMask];
		kDoubleScreenCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Full Size" action:@selector(menuAction:) keyEquivalent:@"f"]; [menu addItem:menuItem];
		if (!soleMenu) [menuItem setKeyEquivalentModifierMask: NSCommandKeyMask|NSAlternateKeyMask];
		kFullScreenCmd = menuItem;
		menuItem = (NSMenuItem *)[NSMenuItem separatorItem]; [menu addItem:menuItem];
		
		aspectMenu = [[NSMenu alloc] initWithTitle:@"Aspect Ratio"];
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Keep" action:@selector(menuAction:) keyEquivalent:@""]; [aspectMenu addItem:menuItem];
		if(vo_keepaspect) [menuItem setState:NSOnState];
		kKeepAspectCmd = menuItem;
		/*
		 menuItem = [[NSMenuItem alloc] initWithTitle:@"Pan-Scan" action:@selector(menuAction:) keyEquivalent:@""]; [aspectMenu addItem:menuItem];
		 if(vo_panscan) [menuItem setState:NSOnState];
		 kPanScanCmd = menuItem; */
		menuItem = (NSMenuItem *)[NSMenuItem separatorItem]; [aspectMenu addItem:menuItem];
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Original" action:@selector(menuAction:) keyEquivalent:@"0"]; [aspectMenu addItem:menuItem];
		kAspectOrgCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"4:3" action:@selector(menuAction:) keyEquivalent:@""]; [aspectMenu addItem:menuItem];
		kAspectFullCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"16:9" action:@selector(menuAction:) keyEquivalent:@""];	[aspectMenu addItem:menuItem];
		kAspectWideCmd = menuItem;
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Aspect Ratio" action:nil keyEquivalent:@""];
		[menuItem setSubmenu:aspectMenu];
		[menu addItem:menuItem];
		[aspectMenu release];
	
		//Add to menubar
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Movie" action:nil keyEquivalent:@""];
		[menuItem setSubmenu:menu];
		if (soleMenu) [[NSApp mainMenu] addItem:menuItem];
		else {
			int wmi;
			if ([NSApp windowsMenu] && ((wmi = [[NSApp mainMenu] indexOfItemWithSubmenu: [NSApp windowsMenu]])>=0))
				[[NSApp mainMenu] insertItem: menuItem atIndex: wmi];
			else
				[[NSApp mainMenu] addItem:menuItem];
		}
	}
	
	if (soleMenu) {
		menu = [[NSMenu alloc] initWithTitle:@"Window"];
	
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Minimize" action:@selector(performMiniaturize:) keyEquivalent:@"m"]; [menu addItem:menuItem];
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Zoom" action:@selector(performZoom:) keyEquivalent:@""]; [menu addItem:menuItem];

		//Add to menubar
		menuItem = [[NSMenuItem alloc] initWithTitle:@"Window" action:nil keyEquivalent:@""];
		[menuItem setSubmenu:menu];
		[[NSApp mainMenu] addItem:menuItem];
		[NSApp setWindowsMenu:menu];
		[menu release];
		[menuItem release];
	}
}

/*
	Menu Action
 */
- (void)menuAction:(id)sender
{
	uint32_t d_width = image_width;
	uint32_t d_height = image_height;
	NSRect frame;
	
	if(sender == kQuitCmd)
	{
		//running = 0;
		/* mplayer_put_key(KEY_ESC); */
	}
	
	if(sender == kHalfScreenCmd)
	{
		if(isFullscreen)
			[self fullscreen:NO];
		
		winSizeMult = 0.5;
		frame.size.width = (d_width*winSizeMult);
		frame.size.height = ((d_width/movie_aspect)*winSizeMult);
		[window setContentSize: frame.size];
		[self reshape];
	}
	if(sender == kNormalScreenCmd)
	{
		if(isFullscreen)
			[self fullscreen:NO];
		
		winSizeMult = 1;
		frame.size.width = d_width;
		frame.size.height = d_width/movie_aspect;
		[window setContentSize: frame.size];
		[self reshape];
	}
	if(sender == kDoubleScreenCmd)
	{
		if(isFullscreen)
			[self fullscreen:NO];
		
		winSizeMult = 2;
		frame.size.width = d_width*winSizeMult;
		frame.size.height = (d_width/movie_aspect)*winSizeMult;
		[window setContentSize: frame.size];
		[self reshape];
	}
	if(sender == kFullScreenCmd)	
		[self fullscreen:YES];

	if(sender == kKeepAspectCmd)
	{
		vo_keepaspect = (!(vo_keepaspect));
		if(vo_keepaspect)
			[kKeepAspectCmd setState:NSOnState];
		else
			[kKeepAspectCmd setState:NSOffState];
			
		[self reshape];
	}

	/*
	if(sender == kPanScanCmd)
	{
		vo_panscan = (!(vo_panscan));
		if(vo_panscan)
			[kPanScanCmd setState:NSOnState];
		else
			[kPanScanCmd setState:NSOffState];
			
		[self panscan];
	}*/
	
	if(sender == kAspectOrgCmd)
	{
		movie_aspect = old_movie_aspect;
		
		if(isFullscreen)
		{
			[self reshape];
		}
		else
		{
			frame.size.width = d_width*winSizeMult;
			frame.size.height = (d_width/movie_aspect)*winSizeMult;
			[window setContentSize: frame.size];
			[self reshape];
		}
	}
	
	if(sender == kAspectFullCmd)
	{
		movie_aspect = 4.0f/3.0f;
		
		if(isFullscreen)
		{
			[self reshape];
		}
		else
		{
			frame.size.width = d_width*winSizeMult;
			frame.size.height = (d_width/movie_aspect)*winSizeMult;
			[window setContentSize: frame.size];
			[self reshape];
		}
	}
		
	if(sender == kAspectWideCmd)
	{
		movie_aspect = 16.0f/9.0f;

		if(isFullscreen)
		{
			[self reshape];
		}
		else
		{
			frame.size.width = d_width*winSizeMult;
			frame.size.height = (d_width/movie_aspect)*winSizeMult;
			[window setContentSize: frame.size];
			[self reshape];
		}
	}
}

/*
	Setup OpenGL
*/
- (void)prepareOpenGL
{
	glEnable(GL_BLEND); 
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glDisable(GL_CULL_FACE);
	[self reshape];
}

/*
	reshape OpenGL viewport
*/ 
- (void)reshape
{
	uint32_t d_width = image_width;
	uint32_t d_height = image_height;
	float aspectX;
	float aspectY;
	int padding = 0;
	
	NSRect frame = [self frame];
	
	glViewport(0, 0, frame.size.width, frame.size.height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, frame.size.width, frame.size.height, 0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	//set texture frame
	if(vo_keepaspect)
	{
		d_height = ((float)d_width/movie_aspect);
		
		aspectX = (float)((float)frame.size.width/(float)d_width);
		aspectY = (float)((float)(frame.size.height)/(float)d_height);
		
		if((d_height*aspectX)>(frame.size.height))
		{
			padding = (frame.size.width - d_width*aspectY)/2;
			textureFrame = NSMakeRect(padding, 0, d_width*aspectY+padding, d_height*aspectY);
		}
		else
		{
			padding = ((frame.size.height) - d_height*aspectX)/2;
			textureFrame = NSMakeRect(0, padding, d_width*aspectX, d_height*aspectX+padding);
		}
	}
	else
	{
		textureFrame = frame;
	}
}

/*
	Render frame
*/ 
- (void) render
{
	int curTime;
	static int lastTime;

	glClear(GL_COLOR_BUFFER_BIT);	
	
	glEnable(CVOpenGLTextureGetTarget(texture));
	glBindTexture(CVOpenGLTextureGetTarget(texture), CVOpenGLTextureGetName(texture));
	
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glTexCoord2f(upperLeft[0], upperLeft[1]); glVertex2i(	textureFrame.origin.x, textureFrame.origin.y);
	glTexCoord2f(lowerLeft[0], lowerLeft[1]); glVertex2i(	textureFrame.origin.x, textureFrame.size.height);
	glTexCoord2f(lowerRight[0], lowerRight[1]); glVertex2i(	textureFrame.size.width, textureFrame.size.height);
	glTexCoord2f(upperRight[0], upperRight[1]); glVertex2i(	textureFrame.size.width, textureFrame.origin.y);
	glEnd();
	glDisable(CVOpenGLTextureGetTarget(texture));
	
	//render resize box
	if(!isFullscreen)
	{
		NSRect frame = [self frame];
		
		glBegin(GL_LINES);
		glColor4f(0.2, 0.2, 0.2, 0.5);
		glVertex2i(frame.size.width-1, frame.size.height-1); glVertex2i(frame.size.width-1, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-5); glVertex2i(frame.size.width-5, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-9); glVertex2i(frame.size.width-9, frame.size.height-1);

		glColor4f(0.4, 0.4, 0.4, 0.5);
		glVertex2i(frame.size.width-1, frame.size.height-2); glVertex2i(frame.size.width-2, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-6); glVertex2i(frame.size.width-6, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-10); glVertex2i(frame.size.width-10, frame.size.height-1);
		
		glColor4f(0.6, 0.6, 0.6, 0.5);
		glVertex2i(frame.size.width-1, frame.size.height-3); glVertex2i(frame.size.width-3, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-7); glVertex2i(frame.size.width-7, frame.size.height-1);
		glVertex2i(frame.size.width-1, frame.size.height-11); glVertex2i(frame.size.width-11, frame.size.height-1);
		glEnd();
	}
	
	glFlush();
	
	//auto hide mouse cursor and futur on-screen control?
	if(isFullscreen && !mouseHide && !isRootwin)
	{
		int curTime = TickCount()/60;
		static int lastTime = 0;
		
		if( ((curTime - lastTime) >= 5) || (lastTime == 0) )
		{
			CGDisplayHideCursor(kCGDirectMainDisplay);
			mouseHide = YES;
			lastTime = curTime;
		}
	}
	
	//update activity every 30 seconds to prevent
	//screensaver from starting up.
	curTime  = TickCount()/60;
	lastTime = 0;
		
	if( ((curTime - lastTime) >= 30) || (lastTime == 0) )
	{
		UpdateSystemActivity(UsrActivity);
		lastTime = curTime;
	}
}

/*
	Create OpenGL texture from current frame & set texco 
*/ 
- (void) setCurrentTexture
{
	CVReturn rerror = kCVReturnSuccess;
	
	rerror = CVOpenGLTextureCacheCreateTextureFromImage (NULL, textureCache,  currentFrameBuffer,  0, &texture);
	if (rerror != kCVReturnSuccess) {
		error("Failed to create OpenGL texture(%d)\n", rerror);
	}

    CVOpenGLTextureGetCleanTexCoords(texture, lowerLeft, lowerRight, upperRight, upperLeft);
}

/*
	redraw win rect
*/ 
- (void) drawRect: (NSRect *) bounds
{
	[self render];
}

/*
	Toggle Fullscreen
*/
- (void) fullscreen: (BOOL) vo_fs
{
	static NSRect old_frame;
	static NSRect old_view_frame;
	BOOL animate = YES;
	
	screen_frame = [[window screen] frame];
		
	//go fullscreen
	if(vo_fs)
	{
		if(!isRootwin)
		{
			SetSystemUIMode( kUIModeAllHidden, kUIOptionAutoShowMenuBar);
			CGDisplayHideCursor(kCGDirectMainDisplay);
			mouseHide = YES;
		}
		
		old_frame = [window frame];	//save main window size & position
		[window setFrame:screen_frame display:YES animate:animate]; //zoom-in window with nice useless sfx
		old_view_frame = [self bounds];
		
		//fix origin for multi screen setup
		screen_frame.origin.x = 0;
		screen_frame.origin.y = 0;
		[self setFrame:screen_frame];
		[self setNeedsDisplay:YES];
		[window setHasShadow:NO];
		isFullscreen = 1;
	}
	else
	{	
		SetSystemUIMode( kUIModeNormal, 0);
		
		isFullscreen = 0;
		CGDisplayShowCursor(kCGDirectMainDisplay);
		mouseHide = NO;

		//revert window to previous setting
		[self setFrame:old_view_frame];
		[self setNeedsDisplay:YES];
		[window setHasShadow:YES];
		[window setFrame:old_frame display:YES animate:animate];//zoom-out window with nice useless sfx
	}
}

- (unsigned int) width { return image_width; }
- (unsigned int) height { return image_height; }
- (unsigned char*) imageBuffer { return image_data; }
- (unsigned int) format { return (image_format==IMGFMT_YUY2)?1:0; }
- (int) viewID { return did; }

/*
	Toggle ontop
*/
- (void) ontop: (BOOL) vo_ontop
{
	if(vo_ontop)
	{
		[window setLevel:NSScreenSaverWindowLevel];
		isOntop = YES;
	}
	else
	{
		[window setLevel:NSNormalWindowLevel];
		isOntop = NO;
	}
}

/*
	Toggle rootwin
 */
- (void) rootwin: (BOOL) vo_rootwin
{
	if(vo_rootwin)
	{
		[window setLevel:CGWindowLevelForKey(kCGDesktopWindowLevelKey)];
		[window orderBack:self];
		isRootwin = YES;
	}
	else
	{
		[window setLevel:NSNormalWindowLevel];
		isRootwin = NO;
	}
}

/*
	Check event for new event
*/ 
- (void) check_events
{
	event = [NSApp nextEventMatchingMask:NSAnyEventMask untilDate:[NSDate dateWithTimeIntervalSinceNow:0.0001] inMode:NSEventTrackingRunLoopMode dequeue:YES];
	[NSApp sendEvent:event];
}

/*
	From NSView, respond to key equivalents.
*/
- (BOOL)performKeyEquivalent:(NSEvent *)theEvent
{
/*
	switch([theEvent keyCode])
    {
		case 0x21: [window setAlphaValue: winAlpha-=0.05]; return YES;
		case 0x1e: [window setAlphaValue: winAlpha+=0.05]; return YES;
    }
	return NO;
 */
}

//	Process key event
- (void) keyDown: (NSEvent *) theEvent
{
	unsigned int key;
	
	switch([theEvent keyCode])
    {
/*
		case 0x34:
		case 0x24: key = KEY_ENTER; break; */
//		case 0x35: running = 0; break;
/*
		case 0x33: key = KEY_BACKSPACE; break;
		case 0x3A: key = KEY_BACKSPACE; break;
		case 0x3B: key = KEY_BACKSPACE; break;
		case 0x38: key = KEY_BACKSPACE; break;
		case 0x7A: key = KEY_F+1; break;
		case 0x78: key = KEY_F+2; break;
		case 0x63: key = KEY_F+3; break;
		case 0x76: key = KEY_F+4; break;
		case 0x60: key = KEY_F+5; break;
		case 0x61: key = KEY_F+6; break;
		case 0x62: key = KEY_F+7; break;
		case 0x64: key = KEY_F+8; break;
		case 0x65: key = KEY_F+9; break;
		case 0x6D: key = KEY_F+10; break;
		case 0x67: key = KEY_F+11; break;
		case 0x6F: key = KEY_F+12; break;
		case 0x72: key = KEY_INSERT; break;
		case 0x75: key = KEY_DELETE; break;
		case 0x73: key = KEY_HOME; break;
		case 0x77: key = KEY_END; break;
		case 0x45: key = '+'; break;
		case 0x4E: key = '-'; break;
		case 0x30: key = KEY_TAB; break;
		case 0x74: key = KEY_PAGE_UP; break;
		case 0x79: key = KEY_PAGE_DOWN; break;  
		case 0x7B: key = KEY_LEFT; break;
		case 0x7C: key = KEY_RIGHT; break;
		case 0x7D: key = KEY_DOWN; break;
		case 0x7E: key = KEY_UP; break;
		case 0x43: key = '*'; break;
		case 0x4B: key = '/'; break;
		case 0x4C: key = KEY_BACKSPACE; break;
		case 0x41: key = KEY_KPDEC; break;
		case 0x52: key = KEY_KP0; break;
		case 0x53: key = KEY_KP1; break;
		case 0x54: key = KEY_KP2; break;
		case 0x55: key = KEY_KP3; break;
		case 0x56: key = KEY_KP4; break;
		case 0x57: key = KEY_KP5; break;
		case 0x58: key = KEY_KP6; break;
		case 0x59: key = KEY_KP7; break;
		case 0x5B: key = KEY_KP8; break;
		case 0x5C: key = KEY_KP9; break;
		default: key = *[[theEvent characters] UTF8String]; break; */
    }
	//mplayer_put_key(key);
}

/*
	Process mouse button event
*/
- (void) mouseMoved: (NSEvent *) theEvent
{
	if(isFullscreen && !isRootwin)
	{
		CGDisplayShowCursor(kCGDirectMainDisplay);
		mouseHide = NO;
	}
}

- (void) mouseDown: (NSEvent *) theEvent
{
	[self mouseEvent: theEvent];
}

- (void) rightMouseDown: (NSEvent *) theEvent
{
	[self mouseEvent: theEvent];
}

- (void) otherMouseDown: (NSEvent *) theEvent
{
	[self mouseEvent: theEvent];
}

- (void) scrollWheel: (NSEvent *) theEvent
{
/*
	if([theEvent deltaY] > 0)
		mplayer_put_key(MOUSE_BTN3);
	else
		mplayer_put_key(MOUSE_BTN4);
 */
}

- (void) mouseEvent: (NSEvent *) theEvent
{
	/*
	switch( [theEvent buttonNumber] )
	{ 
		case 0: mplayer_put_key(MOUSE_BTN0);break;
		case 1: mplayer_put_key(MOUSE_BTN1);break;
		case 2: mplayer_put_key(MOUSE_BTN2);break;
	}*/
}

/*
	NSResponder
*/ 
- (BOOL) acceptsFirstResponder
{
	return YES;
}

- (BOOL) becomeFirstResponder
{
	return YES;
}

- (BOOL) resignFirstResponder
{
	return YES;
}

- (void)windowWillClose:(NSNotification *)aNotification
{
}
@end


// R API

SEXP RVT_create_display(SEXP sw, SEXP sh, SEXP sf) {
	int w = asInteger(sw), h = asInteger(sh), f = asInteger(sf);
	VideoView *view;
	SEXP r;
	
	if (!initialized) preinit("");
	view = [[VideoView alloc] initWithFrame:NSMakeRect(0, 0, 100, 100) pixelFormat:[VideoView defaultPixelFormat]];
	[view configureWidth:w height:h format: (f?IMGFMT_RGB32:IMGFMT_YUY2)];
	r = R_MakeExternalPtr(view, R_NilValue, R_NilValue);
	R_PreserveObject(r); // we require explicit de-allocation
	setAttrib(r, R_ClassSymbol, mkString("videodisplay"));
	return r;
}

SEXP RVT_close_display(SEXP disp) {
	VideoView *view;
	if (TYPEOF(disp)!=EXTPTRSXP || !inherits(disp, "videodisplay")) error("first parameter must be a display");
	view = (VideoView*) EXTPTR_PTR(disp);
	if (!view) error("no video display");
	[[view window] close];
	[view release];
	R_ReleaseObject(disp);
	return R_NilValue;
}

SEXP RVT_get_display_info(SEXP disp) {
	SEXP i;
	VideoView *view;
	int *ii;
	if (TYPEOF(disp)!=EXTPTRSXP || !inherits(disp, "videodisplay")) error("first parameter must be a display");
	view = (VideoView*) EXTPTR_PTR(disp);
	if (!view) error("no video display");
	i = allocVector(INTSXP, 4);
	ii = INTEGER(i);
	ii[0] = [view width];
	ii[1] = [view height];
	ii[2] = [view viewID];
	ii[3] = [view format];
	return i;
}
	
SEXP RVT_update_display_mono(SEXP disp, SEXP m) {
	int image_width, image_height;
	VideoView *view;
	if (TYPEOF(disp)!=EXTPTRSXP || !inherits(disp, "videodisplay")) error("first parameter must be a display");
	view = (VideoView*) EXTPTR_PTR(disp);
	if (!view) error("no video display");
	image_width = [view width];
	image_height = [view height];
	if (isReal(m)) {
		double *d = REAL(m), *ed = d + (image_width*image_height);
		unsigned char *c = [view imageBuffer];
		while (d < ed) {
			unsigned char v = (unsigned char) ((int)(*d*255.0+0.3));
			*(c++) = 0; *(c++) = v; *(c++) = v; *(c++) = v;
			d++;
		}
	}
	if (isInteger(m)) {
		int *d = INTEGER(m), *ed = d + (image_width*image_height);
		unsigned char *c = [view imageBuffer];
		while (d < ed) {
			unsigned char v = (unsigned char) *d;
			*(c++) = 0; *(c++) = v; *(c++) = v; *(c++) = v;
			d++;
		}
	}
	[view setCurrentTexture];
	[view render];
	return R_NilValue;
}

SEXP RVT_update_display_rgb(SEXP disp, SEXP mr, SEXP mg, SEXP mb) {
	int image_width, image_height;
	VideoView *view;
	if (TYPEOF(disp)!=EXTPTRSXP || !inherits(disp, "videodisplay")) error("first parameter must be a display");
	view = (VideoView*) EXTPTR_PTR(disp);
	if (!view) error("no video display");
	image_width = [view width];
	image_height = [view height];
	if (isReal(mr) && isReal(mg) && isReal(mb)) {
		double *dr = REAL(mr), *ed = dr + (image_width*image_height);
		double *dg = REAL(mg), *db = REAL(mb);
		unsigned char *c = [view imageBuffer];
		while (dr < ed) {
			unsigned char rv, gv, bv;
			{ int i = (int)(*(dr++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; rv = (unsigned char) i; }
			{ int i = (int)(*(dg++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; gv = (unsigned char) i; }
			{ int i = (int)(*(db++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; bv = (unsigned char) i; }
			*(c++) = 0; *(c++) = rv; *(c++) = gv; *(c++) = bv;
		}
	} else if (isInteger(mr) && isInteger(mg) && isInteger(mb)) {
		int *dr = INTEGER(mr), *ed = dr + (image_width*image_height);
		int *dg = INTEGER(mg), *db = INTEGER(mb);
		unsigned char *c = [view imageBuffer];
		while (dr < ed) {
			unsigned char rv, gv, bv;
			{ int i = *(dr++); if (i<0) i=0; if (i>255) i=255; rv = (unsigned char) i; }
			{ int i = *(dg++); if (i<0) i=0; if (i>255) i=255; gv = (unsigned char) i; }
			{ int i = *(db++); if (i<0) i=0; if (i>255) i=255; bv = (unsigned char) i; }
			*(c++) = 0; *(c++) = rv; *(c++) = gv; *(c++) = bv;
		}
	} else error("invalid data - all components must be of the same type");
	[view setCurrentTexture];
	[view render];
	return R_NilValue;
}

SEXP RVT_update_display_yuy2(SEXP disp, SEXP my, SEXP mu, SEXP mv) {
	int image_width, image_height;
	int combined;
	VideoView *view;
	if (TYPEOF(disp)!=EXTPTRSXP || !inherits(disp, "videodisplay")) error("first parameter must be a display");
	view = (VideoView*) EXTPTR_PTR(disp);
	if (!view) error("no video display");
	image_width = [view width];
	image_height = [view height];
	combined = (mv==R_NilValue);
	if (isReal(my) && isReal(mu) && combined) {
		double *dy = REAL(my), *ed = dy + (image_width*image_height);
		double *du = REAL(mu);
		unsigned char *c = [view imageBuffer];
		if (LENGTH(my)<image_width*image_height) ed = dy + LENGTH(my);
		if (LENGTH(mu)<LENGTH(my))
			error("dimensions of Y and UV must match");
		ed--; /* we need at least two samples as we go YUYV */
		while (dy < ed) {
			unsigned char yv1, yv2, uv, vv;
			{ int i = (int)(*(dy++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; yv1 = (unsigned char) i; }
			{ int i = (int)((0.5+*(du++))*255.0+0.3); if (i<0) i=0; if (i>255) i=255; uv = (unsigned char) i; }
			{ int i = (int)(*(dy++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; yv2 = (unsigned char) i; }
			{ int i = (int)((0.5+*(du++))*255.0+0.3); if (i<0) i=0; if (i>255) i=255; vv = (unsigned char) i; }
			*(c++) = yv1; *(c++) = uv; *(c++) = yv2; *(c++) = vv;
		}
	} else if (isReal(my) && isReal(mu) && isReal(mv)) {
		double *dy = REAL(my), *ed = dy + (image_width*image_height);
		double *du = REAL(mu), *dv = REAL(mv);
		unsigned char *c = [view imageBuffer];
		if (LENGTH(my)<image_width*image_height) ed = dy + LENGTH(my);
		if (LENGTH(mu) < LENGTH(my)/2 || LENGTH(mv) < LENGTH(my)/2)
			error("dimensions of U and V must be at least half of Y (expand vertically when using 4:1:1)");
		ed--; /* we need at least two samples as we go YUYV */
		while (dy < ed) {
			unsigned char yv1, yv2, uv, vv;
			{ int i = (int)(*(dy++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; yv1 = (unsigned char) i; }
			{ int i = (int)((0.5+*(du++))*255.0+0.3); if (i<0) i=0; if (i>255) i=255; uv = (unsigned char) i; }
			{ int i = (int)(*(dy++)*255.0+0.3); if (i<0) i=0; if (i>255) i=255; yv2 = (unsigned char) i; }
			{ int i = (int)((0.5+*(dv++))*255.0+0.3); if (i<0) i=0; if (i>255) i=255; vv = (unsigned char) i; }
			*(c++) = yv1; *(c++) = uv; *(c++) = yv2; *(c++) = vv;
		}
	} else error("invalid data - all components must be real numbers");
	[view setCurrentTexture];
	[view render];
	return R_NilValue;
}

#endif /* __APPLE__ */
